//
//  TruncatingTableViewString.mm
//
//  Created by Simon Woodside on Sun Mar 16 2003.
//

#import "TruncatingTableViewString.h"

@interface NSString (TruncatingTableViewString_Private)
// Convenience methods for handling attributed strings
- (NSAttributedString *)attributedString:(NSAttributedString *)as
                 stringByAppendingString:(NSAttributedString *)appAS;
- (NSAttributedString *)attributedString:(NSAttributedString *)as
                        substringToIndex:(int)whereToStop;
- (NSAttributedString *)attributedString:(NSAttributedString *)as
                      substringFromIndex:(int)whereToStart;

// Binary search methods
- (NSString *)binarySearchToFitWithIndicatorString:(NSAttributedString *)indstr
                                        truncating:(CHTruncationMode)mode
                                           toWidth:(double)desiredWidth;
- (NSString *)binarySearchToEndFitWithIndicatorString:(NSAttributedString *)indstr
                                              toWidth:(double)desiredWidth;
- (NSString *)binarySearchToMiddleFitWithIndicatorString:(NSAttributedString *)indstr
                                                 toWidth:(double)desiredWidth;
@end

#pragma mark -

@implementation NSString (TruncatingTableViewString)


#pragma mark -

// Truncate self to fit into the table column
// modes are tail: ThisIsAVeryLongStr...
// middle: ThisIsA...ngString
// or head: ...AVeryLongString (not supported yet)
//
// IndicatorString is probably @"..."
// See http://www.cocoadev.com/index.pl?TruncatingStringsInTableView
- (NSString *)truncatedStringWithIndicatorString:(NSString *)istr
                                      truncating:(CHTruncationMode)mode
                              fittingTableColumn:(NSTableColumn *)tc;
{
  if (!istr) { istr = @"NIL"; }

  NSMutableDictionary * attris =
    [[[tc dataCell] attributedStringValue] attributesAtIndex:0 effectiveRange:nil];
  NSAttributedString * attrStr =
    [[NSAttributedString alloc] initWithString:istr attributes:attris];
  double desiredWidth =
    [[tc dataCell] titleRectForBounds:
      NSMakeRect(0, 0, [tc width], [[tc tableView] rowHeight])
      ].size.width;
  desiredWidth -= 5.0; //seems always a bit too big in my outlineView
//  NSLog(@"(TruncatingTableViewString) desiredWidth=%f", desiredWidth);

  if( [self sizeWithAttributes:attris].width <= desiredWidth ) {
//    NSLog(@"(TruncatingTableViewString) end with(as is)=%@", self);
    return self; // string fits as is
  } else {
    NSString * result =
      [self binarySearchToFitWithIndicatorString:attrStr
                                      truncating:(CHTruncationMode)mode
                                         toWidth:desiredWidth];
//    NSLog(@"(TruncatingTableViewString) end with result=%@", result);
    return result;
  }
}

// Public convenience methods to truncate to table width
- (NSString *)endTruncatedStringWithIndicatorString:(NSString *)istr
                                 fittingTableColumn:(NSTableColumn *)tc;
{
  return [self truncatedStringWithIndicatorString:istr
                                       truncating:CHTruncateTail
                               fittingTableColumn:tc];
}
- (NSString *)middleTruncatedStringWithIndicatorString:(NSString *)istr
                                    fittingTableColumn:(NSTableColumn *)tc;
{
  return [self truncatedStringWithIndicatorString:istr
                                       truncating:CHTruncateMiddle
                               fittingTableColumn:tc];
}


#pragma mark -
#pragma mark BINARY SEARCH

- (NSString *)binarySearchToFitWithIndicatorString:(NSAttributedString *)indstr
                                        truncating:(CHTruncationMode)mode
                                           toWidth:(double)desiredWidth;
{
  switch( mode ) {
    case CHTruncateTail:
      return [self binarySearchToEndFitWithIndicatorString:indstr
                                                   toWidth:desiredWidth]; break;
    case CHTruncateMiddle:
      return [self binarySearchToMiddleFitWithIndicatorString:indstr
                                                      toWidth:desiredWidth]; break;
    default:
      return @"UNSUPPORTEDMODE";
  }
}

- (NSString *)binarySearchToEndFitWithIndicatorString:(NSAttributedString *)indstr
                                              toWidth:(double)desiredWidth;
{
  double delta;
  int intDelta;
  int whereToStop;
  double newTestWidth, oldTestWidth;
  NSAttributedString * testString;
  bool haltNow = NO;
  NSAttributedString * result;

  NSAttributedString * attrSelf = [[NSAttributedString alloc] initWithString:self attributes:[indstr attributesAtIndex:0 effectiveRange:nil]];

  // prep
  delta = ceil([attrSelf length] / 2);
  intDelta = (int)delta;
  whereToStop = intDelta - 1; // start at half
  result = [self attributedString:attrSelf substringToIndex:whereToStop];

  // binary search to fit until sub-character precision or fit is reached
  // END
  while( !haltNow )
  {
    if( delta == ceil(delta / 2) ) { haltNow = YES; } // happens at 1.0
    delta = ceil(delta / 2);
    intDelta = (int)delta;

    oldTestWidth = [result size].width + [indstr size].width;
    if( oldTestWidth < desiredWidth )
    {
      testString = [self attributedString:attrSelf
                         substringToIndex:whereToStop + intDelta];
    } else {
      testString = [self attributedString:attrSelf
                         substringToIndex:whereToStop - intDelta];
    }
    newTestWidth = [testString size].width + [indstr size].width;
    //if the new width is closer than the old one
    if( fabs(newTestWidth - desiredWidth) < fabs(oldTestWidth - desiredWidth) )
    {
      //commit the change to result
      result = [testString copy];
      whereToStop = [result length];
    }

    //    NSLog(@"LOOP delta=%f newTestWidth=%f whereToStop=%d result='%@'", delta, newTestWidth, whereToStop, [result string]);
  }
  return [[self attributedString:result stringByAppendingString:indstr] string];
}



//Probably will have SERIOUS problems if the string is very short!!!

- (NSString *)binarySearchToMiddleFitWithIndicatorString:(NSAttributedString *)indstr
                                                 toWidth:(double)desiredWidth;
{
  double delta;
  int intDelta;
  int whereToStop, whereToStart; // left and right offsets (indices)
  double newTestWidth, oldTestWidth;
  bool haltNow = NO;
  NSAttributedString *leftTestString, *rightTestString;
  NSAttributedString *leftResult, *rightResult;

  NSAttributedString * attrSelf = [[NSAttributedString alloc] initWithString:self attributes:[indstr attributesAtIndex:0 effectiveRange:nil]];
  //
  // prep
  //
  delta = ceil([attrSelf length] / 2);
  intDelta = (int)delta;
  whereToStop = intDelta - 1; // start at half
  leftResult = [self attributedString:attrSelf substringToIndex:whereToStop];
  whereToStart = intDelta - 1; //XXXXX maybe duplicate
  rightResult = [self attributedString:attrSelf substringFromIndex:whereToStart];
  leftTestString = [leftResult copy]; rightTestString = [rightResult copy];
  //
  // binary search to fit until sub-character precision or fit is reached
  //
  while( !haltNow )
  {
    if( delta == ceil(delta / 2) ) { haltNow = YES; } // happens at 1.0
    delta = ceil(delta / 2);
    intDelta = (int)delta;

    // find width before and after
    oldTestWidth = [leftResult size].width
      + [indstr size].width + [rightResult size].width;

    if( oldTestWidth > desiredWidth ) {
      intDelta = 0 - intDelta;
    }  // add or subtract?
    leftTestString =
      [self attributedString:attrSelf substringToIndex:(whereToStop+intDelta)];
    rightTestString =
      [self attributedString:attrSelf substringFromIndex:(whereToStart-intDelta)];
    newTestWidth = [leftTestString size].width
      + [indstr size].width + [rightTestString size].width;

    //if the new width is closer than the old one
    if( fabs(newTestWidth - desiredWidth) < fabs(oldTestWidth - desiredWidth) ) {
      //commit the change to result
      leftResult = [leftTestString copy]; whereToStop += intDelta;
      rightResult = [rightTestString copy]; whereToStart -= intDelta;
    }
//    NSLog(@"LOOP delta=%f newTestWidth=%f whereToStop=%d whereToStart=%d doLeftSide=n/a", delta, newTestWidth, whereToStop, whereToStart);//, doLeftSide);
//      NSLog(@"leftResult='%@' rightResult='%@'", [leftResult string], [rightResult string]);
  }
  return [[self attributedString:[self attributedString:leftResult
                                stringByAppendingString:indstr]
         stringByAppendingString:rightResult] string];
}

#pragma mark -
#pragma mark UTILITIES

//result only has attributes from as
- (NSAttributedString *)attributedString:(NSAttributedString *)as
                 stringByAppendingString:(NSAttributedString *)appAS;
{
  return [[NSAttributedString alloc] initWithString:
    [[as string] stringByAppendingString:[appAS string]]
                                         attributes:[as attributesAtIndex:0
                                                           effectiveRange:nil]];
}
//result has all attributes from as
- (NSAttributedString *)attributedString:(NSAttributedString *)as
                        substringToIndex:(int)whereToStop;
{
  return [[NSAttributedString alloc] initWithString:
    [[as string] substringToIndex:whereToStop]
                                         attributes:[as attributesAtIndex:0
                                                           effectiveRange:nil]];
}
//result has all attributes from as
- (NSAttributedString *)attributedString:(NSAttributedString *)as
                      substringFromIndex:(int)whereToStart;
{
  return [[NSAttributedString alloc] initWithString:
    [[as string] substringFromIndex:whereToStart]
                                         attributes:[as attributesAtIndex:0
                                                           effectiveRange:nil]];
}

@end
